#!/usr/bin/env python

"""
High-level classes.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.filesystem import *
from iixr.merging import *
from itertools import islice
from os import mkdir    # index discovery
from os.path import exists

# Constants.

TERM_INTERVAL     = 100
DOCUMENT_INTERVAL = 100
FIELD_INTERVAL    = 100
FLUSH_INTERVAL    = 10000
OPEN_PARTITIONS   = 20

# High-level classes.

class Document:

    "A container of document information."

    def __init__(self, docnum, fields=None):
        self.docnum = docnum
        self.fields = fields or []
        self.terms = {}
        self.field_dict = None

    def add_position(self, term, position):

        """
        Add a position entry for the given 'term', indicating the given
        'position'.
        """

        self.terms.setdefault(term, []).append(position)

    def add_field(self, identifier, value):

        "Add a field having the given 'identifier' and 'value'."

        self.fields.append((identifier, unicode(value))) # convert to string

    def set_fields(self, fields):

        """
        Set the document's 'fields': a list of tuples each containing an integer
        identifier and a string value.
        """

        self.fields = fields

    def _ensure_dict(self):
        if self.field_dict is None:
            self.field_dict = dict(self.fields)

    def keys(self):
        self._ensure_dict()
        return self.field_dict.keys()

    def __getitem__(self, key):
        self._ensure_dict()
        return self.field_dict[key]

class IndexWriter:

    """
    Building term information and writing it to the term and field dictionaries.
    """

    def __init__(self, pathname, interval, doc_interval, field_interval, flush_interval):
        self.pathname = pathname
        self.interval = interval
        self.doc_interval = doc_interval
        self.field_interval = field_interval
        self.flush_interval = flush_interval

        self.dict_partition = get_next_partition(get_term_partitions(self.pathname))
        self.field_dict_partition = get_next_partition(get_field_partitions(self.pathname))

        self.terms = {}
        self.docs = []

        self.doc_counter = 0

    def add_document(self, doc):

        """
        Add the given document 'doc', updating the document counter and flushing
        terms and fields if appropriate.
        """

        for term, positions in doc.terms.items():
            self.terms.setdefault(term, {})[doc.docnum] = positions

        self.docs.append((doc.docnum, doc.fields))

        self.doc_counter += 1
        if self.flush_interval and self.doc_counter >= self.flush_interval:
            self.flush_terms()
            self.flush_fields()
            self.doc_counter = 0

    def get_term_writer(self):

        "Return a term dictionary writer for the current partition."

        return get_term_writer(self.pathname, self.dict_partition, self.interval, self.doc_interval)

    def get_field_writer(self):

        "Return a field dictionary writer for the current partition."

        return get_field_writer(self.pathname, self.field_dict_partition, self.field_interval)

    def flush_terms(self):

        "Flush terms into the current term dictionary partition."

        # Get the terms in order.

        all_terms = self.terms
        terms = all_terms.keys()
        terms.sort()

        dict_writer = self.get_term_writer()

        for term in terms:
            doc_positions = all_terms[term].items()
            dict_writer.write_term_positions(term, doc_positions)

        dict_writer.close()

        self.terms = {}
        self.dict_partition += 1

    def flush_fields(self):

        "Flush fields into the current term dictionary partition."

        # Get the documents in order.

        self.docs.sort()

        field_dict_writer = self.get_field_writer()

        for docnum, fields in self.docs:
            field_dict_writer.write_fields(docnum, fields)

        field_dict_writer.close()

        self.docs = []
        self.field_dict_partition += 1

    def close(self):
        if self.terms or not get_term_partitions(self.pathname):
            self.flush_terms()
        if self.docs or not get_field_partitions(self.pathname):
            self.flush_fields()

class IndexReader:

    "Accessing the term and field dictionaries."

    def __init__(self, pathname):
        self.dict_reader = get_term_reader(pathname, "merged")
        self.field_dict_reader = get_field_reader(pathname, "merged")

    # Sequential access.

    def read_term(self):
        return self.dict_reader.read_term()

    def go_to_term(self, term):
        return self.dict_reader._get_term_and_positions(*self.dict_reader.go_to_term(term))

    # Query access.

    def get_terms(self):
        return self.dict_reader.get_terms()

    def find_terms(self, term):
        return self.dict_reader.find_terms(term)

    def find_positions(self, term):
        return self.dict_reader.find_positions(term)

    def find_common_positions(self, terms):
        return self.dict_reader.find_common_positions(terms)

    def get_frequency(self, term):
        return self.dict_reader.get_frequency(term)

    def get_document_frequency(self, term):
        return self.dict_reader.get_document_frequency(term)

    def get_fields(self, docnum):
        return self.field_dict_reader.get_fields(docnum)

    def get_document(self, docnum):
        return Document(docnum, self.get_fields(docnum))

    def close(self):
        self.dict_reader.close()
        self.field_dict_reader.close()

class Index:

    "An inverted index solution encapsulating the various components."

    def __init__(self, pathname, interval=TERM_INTERVAL, doc_interval=DOCUMENT_INTERVAL, field_interval=FIELD_INTERVAL,
        flush_interval=FLUSH_INTERVAL, open_partitions=OPEN_PARTITIONS):

        self.pathname = pathname
        self.interval = interval
        self.doc_interval = doc_interval
        self.field_interval = field_interval
        self.flush_interval = flush_interval
        self.open_partitions = open_partitions
        self.reader = None
        self.writer = None

    def get_writer(self):

        "Return a writer."

        self._ensure_directory()
        self.writer = IndexWriter(self.pathname, self.interval, self.doc_interval,
            self.field_interval, self.flush_interval)
        return self.writer

    def _ensure_directory(self):
        if not exists(self.pathname):
            mkdir(self.pathname)

    def get_reader(self, partition=0):

        "Return a reader for the index."

        # Ensure that only one partition exists.

        self.merge()
        return self._get_reader(partition)

    def _get_reader(self, partition):

        "Return a reader for the index."

        if not exists(self.pathname):
            raise OSError, "Index path %r does not exist." % self.pathname

        self.reader = IndexReader(self.pathname)
        return self.reader

    def get_term_partitions(self):

        "Return a set of term partition identifiers."

        return get_term_partitions(self.pathname)

    def get_field_partitions(self):

        "Return a set of field partition identifiers."

        return get_field_partitions(self.pathname)

    def merge(self):

        "Merge/optimise index partitions."

        self._merge_terms()
        self._merge_fields()

    def _merge_dictionaries(self, get_partitions, rename_files, remove_files, get_reader, get_writer, get_merger, intervals):

        "Merge term or field dictionaries."

        partitions = get_partitions()

        # Ensure the correct labelling of a single partition.

        if len(partitions) == 1:
            partition = list(partitions)[0]
            if partition != "merged":
                rename_files(self.pathname, partition, "merged")
            return

        # Merge the partitions.

        old_merged_counter = 0

        while len(partitions) > 1:

            if "merged" in partitions:
                rename_files(self.pathname, "merged", "old-merged-%d" % old_merged_counter)
                partitions.remove("merged")
                partitions.add("old-merged-%d" % old_merged_counter)
                old_merged_counter += 1

            # Process only a certain number at once, avoiding resource limits.

            active_partitions = list(islice(partitions, self.open_partitions))

            readers = []
            for partition in active_partitions:
                readers.append(get_reader(self.pathname, partition))

            # Write directly to a dictionary.

            writer = get_writer(self.pathname, "merged", *intervals)
            merger = get_merger(writer, readers)
            merger.merge()
            merger.close()

            # Remove old files.

            for partition in active_partitions:
                remove_files(self.pathname, partition)

            # Acquire the partitions to check their number again.

            partitions = get_partitions()

    def _merge_terms(self):

        "Merge term dictionaries."

        self._merge_dictionaries(self.get_term_partitions, rename_term_files,
            remove_term_files, get_term_reader, get_term_writer,
            TermDictionaryMerger, [self.interval, self.doc_interval])

    def _merge_fields(self):

        "Merge field dictionaries."

        self._merge_dictionaries(self.get_field_partitions, rename_field_files,
            remove_field_files, get_field_reader, get_field_writer,
            FieldDictionaryMerger, [self.field_interval])

    def update(self, other_indexes):

        "Copy the content of the 'other_indexes' into this index and merge."

        self._ensure_directory()

        for i, index in enumerate(other_indexes):
            for partition in index.get_term_partitions():
                copy_term_files(index.pathname, partition, self.pathname, "-added-%d" % i)
            for partition in index.get_field_partitions():
                copy_field_files(index.pathname, partition, self.pathname, "-added-%d" % i)

        self.merge()

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.writer is not None:
            self.writer.close()
            self.writer = None

# vim: tabstop=4 expandtab shiftwidth=4
